# pylint: disable=invalid-name
import os
from robot.api import logger  # noqa: F401 #pylint: disable=unused-import
from robot.utils import is_truthy
import clr
DLL_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'bin', 'TestStack.White.dll')
clr.AddReference('System')
clr.AddReference(DLL_PATH)
from System.Windows.Automation import AutomationElement, ControlType    # noqa: E402
from TestStack.White.UIItems.Finders import SearchCriteria    # noqa: E402
from TestStack.White.UIItems import UIItem    # noqa: E402
from rbfDesktop.keywords import ApplicationKeywords, KeyboardKeywords, WindowKeywords, ScreenshotKeywords, WhiteConfigurationKeywords    # noqa: E402
from rbfDesktop.keywords.items import (ButtonKeywords,
                                         LabelKeywords,
                                         ListKeywords,
                                         ListViewKeywords,
                                         MenuKeywords,
                                         MouseKeywords,
                                         ProgressbarKeywords,
                                         SliderKeywords,
                                         TabKeywords,
                                         TreeKeywords,
                                         TextBoxKeywords,
                                         UiItemKeywords)   # noqa: E402
from rbfDesktop.keywords.robotlibcore import DynamicCore   # noqa: E402
from .version import __version__


STRATEGIES = dict(id={"method": "ByAutomationId"},  # noqa: C408
                  text={"method": "ByText"},
                  index={"method": "Indexed"},
                  help_text={"method": "ByNativeProperty", "property": "HelpTextProperty"},
                  class_name={"method": "ByClassName"},
                  control_type={"method": "ByControlType"})


class rbfDesktop(DynamicCore):
    """rbfDesktop is a Robot Framework library for automating Windows GUI.

    It is a wrapper for [https://github.com/TestStack/White | TestStack.White] automation framework, which is based on
    [https://docs.microsoft.com/en-us/windows/desktop/WinAuto/entry-uiauto-win32 | Microsoft UI Automation API] (UIA).

    = Applications and windows =
    To interact with UI items, the correct application and window must be attached to rbfDesktop.

    When application is started with `Launch Application`, the keyword also attaches the application to rbfDesktop.
    Attaching a running application is done with `Attach Application By Name` or `Attach Application By Id`.

    Once the application is attached, the window to interact with is attached with `Attach Window`.

    Examples:
    | # Launch application, no separate step for attaching application needed | |
    | `Launch Application` | C:/myApplication.exe |
    | `Attach Window`      | Main window |
    | | |
    | # Switch to an application that is already running | |
    | `Attach Application By Name` | calc1 |
    | `Attach Window`              | Calculator |

    = UI items =
    rbfDesktop uses the same names for UI items (=controls) as White.
    See [https://teststackwhite.readthedocs.io/en/latest/UIItems | White's documentation] for details about mapping
    UIA control types to White's UI item classes.

    For example, the UIA control type ``Text`` maps to the ``Label`` class in White (e.g. in rbfDesktop's keyword `Verify Label`).

    == Item locators ==
    Keywords that access UI items (e.g. `Click Button`) use a ``locator`` argument.
    The locator consists of a locator prefix that specifies the search criteria, and the locator value.

    Locator syntax is ``prefix:value``.
    The following locator prefixes are available:

    | = Prefix =        | = Description =                    |
    | id (or no prefix) | Search by AutomationID. If no prefix is given, the item is searched by AutomationID by default. |
    | text              | Search by exact item text or name. |
    | index             | Search by item index.              |
    | help_text         | Search by HelpTextProperty.        |
    | class_name        | Search by class name.              |
    | control_type      | Search by control type.            |

    Examples:
    | `Click Button` | myButton         | # clicks button by its AutomationID |
    | `Click Button` | id:myButton      | # clicks button by its AutomationID |
    | `Click Button` | text:Click here! | # clicks button by the button text  |
    | `Click Button` | index:2          | # clicks button whose index is 2    |

    *Note:* Old locator syntax ``prefix=value`` is also valid but it is recommended to use the ``prefix:value`` syntax
    since the old syntax *will be deprecated* in the future.

    == Item object as a locator ==
    It is also possible to use an item object reference as the ``locator`` value.
    An item object can be obtained with e.g. `Get Item` or `Get Items` keywords.

    The need to use an item object reference can arise for instance when multiple items match the same locator
    and one of the items is selected for further action.
    When using an item object, the action on the item can be executed regardless of the window it is in,
    i.e. the window where the item is located does not necessarily need to be attached.
    However, this does not change the attached window and the operation continues in the attached window after action on
    the referred item is complete.

    Example using item object:
    | @{my_buttons}= | `Get Items`         | class_name:MyButtonClass |
    | `Click Button` | ${my_buttons[2]}    | # clicks button object at index 2 of the list |

    = Workflow example =
    | ***** Variables *****   | | | |
    | ${TEST APPLICATION}     | C:/path/to/my_application.exe | | |
    | | | | |
    | ***** Settings *****    | | | |
    | Library | rbfDesktop  | | |
    | | | | |
    | ***** Test Cases *****  | | | |
    | Small Example           | | | |
    | | Launch Application    | ${TEST APPLICATION} | |
    | | Attach Window         | Window Title        | |
    | | Button Text Should Be | my_button           | press this button |
    | | Click Button          | my_button | |
    | | Close Application     | | |

    """

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = __version__
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, screenshot_dir=None):
        """rbfDesktop can be imported with an optional argument ``screenshot_dir``.

        ``screenshot_dir`` is the directory where screenshots taken by rbfDesktop are saved.
        If the given directory does not already exist, it will be created when the first screenshot is taken.
        The directory can also be set at runtime with `Set Screenshot Directory`.
        If the argument is not given, the default location for screenshots is the output directory of the Robot run,
        i.e. the directory where output and log files are generated.
        """
        self.app = None
        self.window = None
        self.screenshooter = None
        self.ROBOT_LIBRARY_LISTENER = self  # pylint: disable=invalid-name
        self.screenshots_enabled = True
        self.libraries = [ApplicationKeywords(self),
                          ButtonKeywords(self),
                          KeyboardKeywords(self),
                          LabelKeywords(self),
                          ListKeywords(self),
                          ListViewKeywords(self),
                          MenuKeywords(self),
                          MouseKeywords(self),
                          ProgressbarKeywords(self),
                          SliderKeywords(self),
                          TabKeywords(self),
                          WhiteConfigurationKeywords(self),
                          TextBoxKeywords(self),
                          TreeKeywords(self),
                          UiItemKeywords(self),
                          WindowKeywords(self),
                          ScreenshotKeywords(self, screenshot_dir)]
        DynamicCore.__init__(self, self.libraries)

    def _get_typed_item_by_locator(self, item_type, locator):
        if isinstance(locator, UIItem):
            if not isinstance(locator, item_type):
                raise TypeError("Item object was not of the expected type")
            return locator
        search_criteria = self._get_search_criteria(locator)
        return self.window.Get[item_type](search_criteria)

    def _get_item_by_locator(self, locator):
        if isinstance(locator, UIItem):
            return locator
        search_criteria = self._get_search_criteria(locator)
        return self.window.Get(search_criteria)

    def _get_multiple_items_by_locator(self, locator):
        search_criteria = self._get_search_criteria(locator)
        return self.window.GetMultiple(search_criteria)

    def _get_search_criteria(self, locator):
        search_strategy, locator_value = self._parse_locator(locator)
        if search_strategy == "index":
            locator_value = int(locator_value)

        try:
            search_method = STRATEGIES[search_strategy]["method"]
        except KeyError:
            raise ValueError("'{}' is not a valid locator prefix".format(search_strategy))

        if search_method == "ByNativeProperty":
            property_name = STRATEGIES[search_strategy]["property"]
            property_name = getattr(AutomationElement, property_name)
            search_params = (property_name, locator_value)
        else:
            if search_method == "ByControlType":
                locator_value = getattr(ControlType, locator_value)
            search_params = (locator_value,)

        method = getattr(SearchCriteria, search_method)
        return method(*search_params)

    def _parse_locator(self, locator):
        if "=" not in locator and ":" not in locator:
            locator = "id:" + locator
        idx = self._get_locator_delimiter_index(locator)
        return locator[:idx], locator[idx + 1:]

    @staticmethod
    def _get_locator_delimiter_index(locator):
        if "=" not in locator:
            return locator.index(":")
        if ":" not in locator:
            return locator.index("=")
        return min(locator.index(":"), locator.index("="))

    def _end_keyword(self, name, attrs):  # pylint: disable=unused-argument
        if attrs['status'] == 'FAIL':
            if self.screenshots_enabled:
                self.screenshooter.take_desktop_screenshot()

    @staticmethod
    def _contains_string_value(expected, actual, case_sensitive=True):
        case_sensitive = is_truthy(case_sensitive)
        expected_value = expected if case_sensitive else expected.upper()
        actual_value = actual if case_sensitive else actual.upper()
        if expected_value not in actual_value:
            raise AssertionError("Expected value {} not found in {}".format(expected, actual))

    @staticmethod
    def _verify_string_value(expected, actual, case_sensitive=True):
        case_sensitive = is_truthy(case_sensitive)
        expected_value = expected if case_sensitive else expected.upper()
        actual_value = actual if case_sensitive else actual.upper()
        if expected_value != actual_value:
            raise AssertionError("Expected value {}, but found {}".format(expected, actual))

    @staticmethod
    def _verify_value(expected, actual):
        if expected != actual:
            raise AssertionError("Expected value {}, but found {}".format(expected, actual))



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import types
import requests

from selenium import webdriver

from rbfWEB.base import keyword, LibraryComponent
from rbfWEB.locators import WindowManager
from rbfWEB.utils import (is_truthy, is_noney, secs_to_timestr,
                                   timestr_to_secs)

from .webdrivertools import WebDriverCreator
from .waiting import WaitingKeywords
from .element import ElementKeywords

from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys


class NettbankKeywords(LibraryComponent):

    def __init__(self, ctx):
        LibraryComponent.__init__(self, ctx)
        self._window_manager = WindowManager(ctx)

       
    @keyword
    def DNB_get_OTP_for_Uten_BankID(self, serialNumber, url=None):
        
        """Generates a new OTP for Uten Bank ID Authentication for GO3 & DP250 devices for the given ``serialNumber``. 
        
        ``url`` & ``alias`` are Optional 
        
        Examples:
        | `DNBGet Otp With Gogo` | 24299558 |
        | `Get Otp With Gogo` | 24299558 | http://e41ap027.erf01.net:8080/getOtp/jsp/getCode.servlet |
        | `Get Otp With Gogo` | 24299558 | http://e41ap027.erf01.net:8080/getOtp/jsp/getCode.servlet | getOTP |    
        
        """
        if url is None:
            formattedURL="http://e41ap027.erf01.net:8080/getOtp/jsp/getCode.servlet"+"?"+"serialNumber="+serialNumber+"&GoGo=GoGo"
        if url is not None:
            formattedURL=url+"?"+"serialNumber="+serialNumber+"&GoGo=GoGo"
        # browser='headlesschrome'
        # self.info("Opening browser '%s' to base url '%s'." % (browser, formattedURL))
        # driver = self._make_driver(browser)
        # try:
            # driver.get(formattedURL)
        # except Exception:
            # self.ctx.register_driver(driver, alias)
            # self.debug("Opened browser with session id %s but failed "
                       # "to open url '%s'." % (driver.session_id, formattedURL))
            # raise
        # self.debug('Opened browser with session id %s.' % driver.session_id)
        # self.ctx.register_driver(driver, alias)
        # self.generatedText=self.find_element('//*[text()[contains(., \'OTP Code\')]]').text
        # self.otp=self._get_substring(self.generatedText, -6)
        # self.formattedOTP=self._format_OTP(self.otp)
        # self.close_browser()
        # return self.formattedOTP
        
        self.info("Making a POST call to '%s'." % (formattedURL))
        response = requests.post(url = formattedURL)
        responseString = str(response.text)
        self.otp = responseString.partition("OTP Code ")[2][:6]
        self.formattedOTP=self._format_OTP(self.otp)
        return self.formattedOTP
        
    @keyword
    def open_browser(self, url, browser='firefox', alias=None,
                     remote_url=False, desired_capabilities=None,
                     ff_profile_dir=None):
        """Opens a new browser instance to the given ``url``.

        The ``browser`` argument specifies which browser to use, and the
        supported browser are listed in the table below. The browser names
        are case-insensitive and some browsers have multiple supported names.

        |    = Browser =    |        = Name(s) =       |
        | Firefox           | firefox, ff              |
        | Google Chrome     | googlechrome, chrome, gc |
        | Headless Firefox  | headlessfirefox          |
        | Headless Chrome   | headlesschrome           |
        | Internet Explorer | internetexplorer, ie     |
        | Edge              | edge                     |
        | Safari            | safari                   |
        | Opera             | opera                    |
        | Android           | android                  |
        | Iphone            | iphone                   |
        | PhantomJS         | phantomjs                |
        | HTMLUnit          | htmlunit                 |
        | HTMLUnit with Javascript | htmlunitwithjs    |

        To be able to actually use one of these browsers, you need to have
        a matching Selenium browser driver available. Headless Firefox and
        Headless Chrome require Selenium 3.8.0 or newer.

        Optional ``alias`` is an alias given for this browser instance and
        it can be used for switching between browsers. An alternative
        approach for switching is using an index returned by this keyword.
        These indices start from 1, are incremented when new browsers are
        opened, and reset back to 1 when `Close All Browsers` is called.
        See `Switch Browser` for more information and examples.

        Optional ``remote_url`` is the URL for a
        [https://github.com/SeleniumHQ/selenium/wiki/Grid2|Selenium Grid].

        Optional ``desired_capabilities`` can be used to configure, for example,
        logging preferences for a browser or a browser and operating system
        when using [http://saucelabs.com|Sauce Labs]. Desired capabilities can
        be given either as a Python dictionary or as a string in format
        ``key1:value1,key2:value2``.
        [https://github.com/SeleniumHQ/selenium/wiki/DesiredCapabilities|
        Selenium documentation] lists possible capabilities that can be
        enabled.

        Optional ``ff_profile_dir`` is the path to the Firefox profile
        directory if you wish to overwrite the default profile Selenium
        uses. 

        Examples:
        | `Open Browser` | http://example.com | Chrome  |
        | `Open Browser` | http://example.com | Firefox | alias=Firefox |
        | `Open Browser` | http://example.com | Edge    | remote_url=http://127.0.0.1:4444/wd/hub |

        If the provided configuration options are not enough, it is possible
        to use `Create Webdriver` to customize browser initialization even
        more.

        Applying ``desired_capabilities`` argument also for local browser is
        
        """
        if is_truthy(remote_url):
            self.info("Opening browser '%s' to base url '%s' through "
                      "remote server at '%s'." % (browser, url, remote_url))
        else:
            self.info("Opening browser '%s' to base url '%s'." % (browser, url))
        driver = self._make_driver(browser, desired_capabilities,
                                   ff_profile_dir, remote_url)
        try:
            driver.get(url)
        except Exception:
            self.ctx.register_driver(driver, alias)
            self.debug("Opened browser with session id %s but failed "
                       "to open url '%s'." % (driver.session_id, url))
            raise
        self.debug('Opened browser with session id %s.' % driver.session_id)
        return self.ctx.register_driver(driver, alias)


    
    def _wait_until_page_contains(self, text, timeout=None, error=None):
        """Waits until ``text`` appears on current page.

        Fails if ``timeout`` expires before the text appears. See
        the `Timeouts` section for more information about using timeouts
        and their default value.

        ``error`` can be used to override the default error message.
        """
        self._wait_until(lambda: self.is_text_present(text),
                         "Text '%s' did not appear in <TIMEOUT>." % text,
                         timeout, error)
    
    
    def _wait_until(self, condition, error, timeout=None, custom_error=None):
        timeout = self.get_timeout(timeout)
        if is_noney(custom_error):
            error = error.replace('<TIMEOUT>', secs_to_timestr(timeout))
        else:
            error = custom_error
        self._wait_until_worker(condition, timeout, error)
    
    
    def _wait_until_worker(self, condition, timeout, error):
        max_time = time.time() + timeout
        not_found = None
        while time.time() < max_time:
            try:
                if condition():
                    return
            except ElementNotFound as err:
                not_found = str(err)
            except StaleElementReferenceException as err:
                self.info('Suppressing StaleElementReferenceException from Selenium.')
                not_found = err
            else:
                not_found = None
            time.sleep(0.2)
        raise AssertionError(not_found or error)
        
    
    def _make_driver(self, browser, desired_capabilities=None,
                     profile_dir=None, remote=None):
        driver = WebDriverCreator(self.log_dir).create_driver(
            browser=browser, desired_capabilities=desired_capabilities,
            remote_url=remote, profile_dir=profile_dir)
        driver.set_script_timeout(self.ctx.timeout)
        driver.implicitly_wait(self.ctx.implicit_wait)
        if self.ctx.speed:
            self._monkey_patch_speed(driver)
        return driver

    def _monkey_patch_speed(self, driver):
        def execute(self, driver_command, params=None):
            result = self._base_execute(driver_command, params)
            speed = self._speed if hasattr(self, '_speed') else 0.0
            if speed > 0:
                time.sleep(speed)
            return result
        if not hasattr(driver, '_base_execute'):
            driver._base_execute = driver.execute
            driver.execute = types.MethodType(execute, driver)
        driver._speed = self.ctx.speed
        
    def _get_substring(self, string, start, end=None):
        """Returns a substring from ``start`` index to ``end`` index.

        The ``start`` index is inclusive and ``end`` is exclusive.
        Indexing starts from 0, and it is possible to use
        negative indices to refer to characters from the end.

        Examples:
        | ${ignore first} = | Get Substring | ${string} | 1  |    |
        | ${ignore last} =  | Get Substring | ${string} |    | -1 |
        | ${5th to 10th} =  | Get Substring | ${string} | 4  | 10 |
        | ${first two} =    | Get Substring | ${string} |    | 1  |
        | ${last two} =     | Get Substring | ${string} | -2 |    |
        """
        start = self._convert_to_index(start, 'start')
        end = self._convert_to_index(end, 'end')
        return string[start:end]
    
    def _convert_to_index(self, value, name):
        if value == '':
            return 0
        if value is None:
            return None
        return self._convert_to_integer(value, name)

    
    def _convert_to_integer(self, value, name):
        try:
            return int(value)
        except ValueError:
            raise ValueError("Cannot convert '%s' argument '%s' to an integer."
                             % (name, value))
    
    

    def _format_OTP(self, otp):
        formattedOTP = ''
        for i in otp:
            if i in '0123456789':
                formattedOTP+=i
    
        while (len(formattedOTP)<6):
            num = '0'+formattedOTP        
        return formattedOTP   
    
    
    def close_browser(self):
        """Closes the current browser."""
        if self.drivers.current:
            self.debug('Closing browser with session id {}.'
                       .format(self.driver.session_id))
            self.drivers.close()




from TestStack.White.UIItems import UIItem  # noqa: F401 #pylint: disable=unused-import
from rbfDesktop.keywords.librarycomponent import LibraryComponent
from rbfDesktop.keywords.robotlibcore import keyword
from rbfDesktop.utils.click import Clicks


class UiItemKeywords(LibraryComponent):
    @keyword
    def click_item(self, locator, x_offset=0, y_offset=0):
        """Clicks an item.

        ``locator`` is the locator of the item or object of an item.
        Locator syntax is explained in `Item locators`.

        Optional arguments ``x_offset`` and ``y_offset`` can be used to define the coordinates to click at,
        relative to the center of the item.
        """
        item = self.state._get_item_by_locator(locator)
        Clicks.click(item, x_offset, y_offset)

    @keyword
    def right_click_item(self, locator, x_offset=0, y_offset=0):
        """Right clicks an item.

        ``locator`` is the locator of the item or object of an item.
        Locator syntax is explained in `Item locators`.

        Optional arguments ``x_offset`` and ``y_offset`` can be used to define the coordinates to click at,
        relative to the center of the item.
        """
        item = self.state._get_item_by_locator(locator)
        Clicks.right_click(item, x_offset, y_offset)

    @keyword
    def double_click_item(self, locator, x_offset=0, y_offset=0):
        """Double clicks an item.

        ``locator`` is the locator of the item or object of an item.
        Locator syntax is explained in `Item locators`.

        Optional arguments ``x_offset`` and ``y_offset`` can be used to define the coordinates to click at,
        relative to the center of the item.
        """
        item = self.state._get_item_by_locator(locator)
        Clicks.double_click(item, x_offset, y_offset)

    @keyword
    def get_items(self, locator):
        """Returns a list of items that match the given ``locator``.

        Locator syntax is explained in `Item locators`.
        """
        return self.state._get_multiple_items_by_locator(locator)

    @keyword
    def get_item(self, locator):
        """Returns the first item that matches the given locator.

        ``locator`` is the locator of the item.
        Locator syntax is explained in `Item locators`.
        """
        return self.state._get_item_by_locator(locator)

    @keyword
    def item_should_be_enabled(self, locator):
        """Verifies that an item is enabled.

        ``locator`` is the locator of the item.
        Locator syntax is explained in `Item locators`.
        """
        item = self.state._get_item_by_locator(locator)
        if not item.Enabled:
            raise AssertionError("Expected item with locator '{}' to be enabled but found disabled".format(locator))

    @keyword
    def item_should_be_disabled(self, locator):
        """Verifies that an item is disabled.

        ``locator`` is the locator of the item.
        Locator syntax is explained in `Item locators`.
        """
        item = self.state._get_item_by_locator(locator)
        if item.Enabled:
            raise AssertionError("Expected item with locator '{}' to be disabled but found enabled".format(locator))



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings

from robot.utils import ConnectionCache
from selenium import webdriver

from rbfWEB.utils import is_falsy, is_truthy, SELENIUM_VERSION


class WebDriverCreator(object):

    browser_names = {
        'googlechrome': "chrome",
        'gc': "chrome",
        'chrome': "chrome",
        'headlesschrome': 'headless_chrome',
        'ff': 'firefox',
        'firefox': 'firefox',
        'headlessfirefox': 'headless_firefox',
        'ie': 'ie',
        'internetexplorer': 'ie',
        'edge': 'edge',
        'opera': 'opera',
        'safari': 'safari',
        'phantomjs': 'phantomjs',
        'htmlunit': 'htmlunit',
        'htmlunitwithjs': 'htmlunit_with_js',
        'android': 'android',
        'iphone': 'iphone'
    }

    def __init__(self, log_dir):
        self.log_dir = log_dir

    def create_driver(self, browser, desired_capabilities, remote_url,
                      profile_dir=None):
        creation_method = self._get_creator_method(browser)
        desired_capabilities = self._parse_capabilities(desired_capabilities, browser)
        if (creation_method == self.create_firefox
                or creation_method == self.create_headless_firefox):
            return creation_method(desired_capabilities, remote_url,
                                   profile_dir)
        return creation_method(desired_capabilities, remote_url)

    def _get_creator_method(self, browser):
        browser = browser.lower().replace(' ', '')
        if browser in self.browser_names:
            return getattr(self, 'create_{}'.format(self.browser_names[browser]))
        raise ValueError('{} is not a supported browser.'.format(browser))

    def _parse_capabilities(self, capabilities, browser=None):
        if is_falsy(capabilities):
            return {}
        if not isinstance(capabilities, dict):
            capabilities = self._string_to_dict(capabilities)
        browser_alias = {'googlechrome': "chrome", 'gc': "chrome",
                         'headlesschrome': 'chrome', 'ff': 'firefox',
                         'headlessfirefox': 'firefox',
                         'internetexplorer': 'ie'}
        browser = browser_alias.get(browser, browser)
        if browser in ['ie', 'firefox', 'edge']:
            return {'capabilities': capabilities}
        return {'desired_capabilities': capabilities}

    def _string_to_dict(self, capabilities):
        desired_capabilities = {}
        for part in capabilities.split(','):
            key, value = part.split(':')
            desired_capabilities[key.strip()] = value.strip()
        return desired_capabilities

    def create_chrome(self, desired_capabilities, remote_url, options=None):
        if is_truthy(remote_url):
            if not desired_capabilities:
                desired_capabilities = {'desired_capabilities': webdriver.DesiredCapabilities.CHROME.copy()}
            return self._remote(desired_capabilities, remote_url, options=options)
        if SELENIUM_VERSION.major >= 3 and SELENIUM_VERSION.minor >= 8:
            return webdriver.Chrome(options=options, **desired_capabilities)
        return webdriver.Chrome(**desired_capabilities)

    def create_headless_chrome(self, desired_capabilities, remote_url):
        if SELENIUM_VERSION.major >= 3 and SELENIUM_VERSION.minor >= 8:
            options = webdriver.ChromeOptions()
            options.set_headless()
        else:
            options = None
        return self.create_chrome(desired_capabilities, remote_url, options)

    def create_firefox(self, desired_capabilities, remote_url, ff_profile_dir,
                       options=None):
        profile = self._get_ff_profile(ff_profile_dir)
        if is_truthy(remote_url):
            if not desired_capabilities:
                desired_capabilities = {'desired_capabilities': webdriver.DesiredCapabilities.FIREFOX.copy()}
            return self._remote(desired_capabilities, remote_url,
                                profile, options)
        desired_capabilities.update(self._geckodriver_log)
        if SELENIUM_VERSION.major >= 3 and SELENIUM_VERSION.minor >= 8:
            return webdriver.Firefox(options=options, firefox_profile=profile,
                                     **desired_capabilities)
        return webdriver.Firefox(firefox_profile=profile,
                                 **desired_capabilities)

    def _get_ff_profile(self, ff_profile_dir):
        if is_falsy(ff_profile_dir):
            return webdriver.FirefoxProfile()
        return webdriver.FirefoxProfile(ff_profile_dir)

    @property
    def _geckodriver_log(self):
        if SELENIUM_VERSION.major >= 3:
            return {'log_path': os.path.join(self.log_dir, 'geckodriver.log')}
        return {}

    def create_headless_firefox(self, desired_capabilities, remote_url,
                                ff_profile_dir):
        if SELENIUM_VERSION.major >= 3 and SELENIUM_VERSION.minor >= 8:
            options = webdriver.FirefoxOptions()
            options.set_headless()
        else:
            options = None
        return self.create_firefox(desired_capabilities, remote_url,
                                   ff_profile_dir, options)

    def create_ie(self, desired_capabilities, remote_url):
        if is_truthy(remote_url):
            if not desired_capabilities:
                ie = webdriver.DesiredCapabilities.INTERNETEXPLORER.copy()
                desired_capabilities = {'desired_capabilities': ie}
            return self._remote(desired_capabilities, remote_url)
        return webdriver.Ie(**desired_capabilities)

    def create_edge(self, desired_capabilities, remote_url):
        if is_truthy(remote_url):
            if not desired_capabilities:
                edge = webdriver.DesiredCapabilities.EDGE.copy()
                desired_capabilities = {'desired_capabilities': edge}
            return self._remote(desired_capabilities, remote_url)
        return webdriver.Edge(**desired_capabilities)

    def create_opera(self, desired_capabilities, remote_url):
        if is_truthy(remote_url):
            if not desired_capabilities:
                opera = webdriver.DesiredCapabilities.OPERA.copy()
                desired_capabilities = {'desired_capabilities': opera}
            return self._remote(desired_capabilities, remote_url)
        return webdriver.Opera(**desired_capabilities)

    def create_safari(self, desired_capabilities, remote_url):
        if is_truthy(remote_url):
            if not desired_capabilities:
                caps = webdriver.DesiredCapabilities.SAFARI.copy()
                desired_capabilities = {'desired_capabilities': caps}
            return self._remote(desired_capabilities, remote_url)
        return webdriver.Safari(**desired_capabilities)

    def create_phantomjs(self, desired_capabilities, remote_url):
        warnings.warn('rbfWEB support for PhantomJS has been deprecated, '
                      'please use headlesschrome or headlessfirefox instead.')
        if is_truthy(remote_url):
            if not desired_capabilities:
                caps = webdriver.DesiredCapabilities.PHANTOMJS.copy()
                desired_capabilities = {'desired_capabilities': caps}
            return self._remote(desired_capabilities, remote_url)
        return webdriver.PhantomJS(**desired_capabilities)

    def create_htmlunit(self, desired_capabilities, remote_url):
        if not desired_capabilities:
            desired_capabilities['desired_capabilities'] = webdriver.DesiredCapabilities.HTMLUNIT
        return self._remote(desired_capabilities, remote_url)

    def create_htmlunit_with_js(self, desired_capabilities, remote_url):
        if not desired_capabilities:
            desired_capabilities['desired_capabilities'] = webdriver.DesiredCapabilities.HTMLUNITWITHJS
        return self._remote(desired_capabilities, remote_url)

    def create_android(self, desired_capabilities, remote_url):
        if not desired_capabilities:
            desired_capabilities['desired_capabilities'] = webdriver.DesiredCapabilities.ANDROID
        return self._remote(desired_capabilities, remote_url)

    def create_iphone(self, desired_capabilities, remote_url):
        if not desired_capabilities:
            desired_capabilities['desired_capabilities'] = webdriver.DesiredCapabilities.IPHONE
        return self._remote(desired_capabilities, remote_url)

    def _remote(self, desired_capabilities, remote_url,
                profile_dir=None, options=None):
        remote_url = str(remote_url)
        if 'capabilities' in desired_capabilities:
            desired_capabilities['desired_capabilities'] = desired_capabilities.pop('capabilities')
        if SELENIUM_VERSION.major >= 3 and SELENIUM_VERSION.minor >= 8:
            return webdriver.Remote(command_executor=remote_url,
                                    browser_profile=profile_dir, options=options,
                                    **desired_capabilities)
        return webdriver.Remote(command_executor=remote_url,
                                browser_profile=profile_dir,
                                **desired_capabilities)


class WebDriverCache(ConnectionCache):

    def __init__(self):
        ConnectionCache.__init__(self, no_current_msg='No current browser')
        self._closed = set()

    @property
    def drivers(self):
        return self._connections

    @property
    def active_drivers(self):
        open_drivers = []
        for driver in self._connections:
            if driver not in self._closed:
                open_drivers.append(driver)
        return open_drivers

    def close(self):
        if self.current:
            driver = self.current
            driver.quit()
            self.current = self._no_current
            self._closed.add(driver)

    def close_all(self):
        for driver in self._connections:
            if driver not in self._closed:
                driver.quit()
        self.empty_cache()
        return self.current



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import namedtuple

from selenium.common.exceptions import (NoSuchWindowException,
                                        WebDriverException)

from rbfWEB.base import ContextAware
from rbfWEB.errors import WindowNotFound
from rbfWEB.utils import is_string


WindowInfo = namedtuple('WindowInfo', 'handle, id, name, title, url')


class WindowManager(ContextAware):

    def __init__(self, ctx):
        ContextAware.__init__(self, ctx)
        self._strategies = {
            'title': self._select_by_title,
            'name': self._select_by_name,
            'url': self._select_by_url,
            'default': self._select_by_default
        }

    def get_window_infos(self):
        infos = []
        try:
            starting_handle = self.driver.current_window_handle
        except NoSuchWindowException:
            starting_handle = None
        try:
            for handle in self.driver.window_handles:
                self.driver.switch_to.window(handle)
                infos.append(self._get_current_window_info())
        finally:
            if starting_handle:
                self.driver.switch_to.window(starting_handle)
        return infos

    def select(self, locator, timeout=0):
        while True:
            try:
                return self._select(locator)
            except WindowNotFound:
                if time.time() > timeout:
                    raise
                time.sleep(0.1)

    def _select(self, locator):
        if not is_string(locator):
            self._select_by_excludes(locator)
        elif locator.upper() == 'CURRENT':
            pass
        elif locator.upper() == 'MAIN':
            self._select_main_window()
        elif locator.upper() == 'NEW':
            self._select_by_last_index()
        else:
            strategy, locator = self._parse_locator(locator)
            self._strategies[strategy](locator)

    def _parse_locator(self, locator):
        index = self._get_locator_separator_index(locator)
        if index != -1:
            prefix = locator[:index].strip()
            if prefix in self._strategies:
                return prefix, locator[index + 1:].lstrip()
        return 'default', locator

    def _get_locator_separator_index(self, locator):
        if '=' not in locator:
            return locator.find(':')
        if ':' not in locator:
            return locator.find('=')
        return min(locator.find('='), locator.find(':'))

    def _select_by_title(self, title):
        self._select_matching(
            lambda window_info: window_info.title == title,
            "Unable to locate window with title '%s'." % title
        )

    def _select_by_name(self, name):
        self._select_matching(
            lambda window_info: window_info.name == name,
            "Unable to locate window with name '%s'." % name
        )

    def _select_by_url(self, url):
        self._select_matching(
            lambda window_info: window_info.url == url,
            "Unable to locate window with URL '%s'." % url
        )

    def _select_main_window(self):
        handles = self.driver.window_handles
        self.driver.switch_to.window(handles[0])

    def _select_by_default(self, criteria):
        try:
            starting_handle = self.driver.current_window_handle
        except NoSuchWindowException:
            starting_handle = None
        for handle in self.driver.window_handles:
            self.driver.switch_to.window(handle)
            if criteria == handle:
                return
            for item in self._get_current_window_info()[2:4]:
                if item == criteria:
                    return
        if starting_handle:
            self.driver.switch_to.window(starting_handle)
        raise WindowNotFound("No window matching handle, name, title or URL "
                             "'%s' found." % criteria)

    def _select_by_last_index(self):
        handles = self.driver.window_handles
        if handles[-1] == self.driver.current_window_handle:
            raise WindowNotFound('Window with last index is same as '
                                 'the current window.')
        self.driver.switch_to.window(handles[-1])

    def _select_by_excludes(self, excludes):
        for handle in self.driver.window_handles:
            if handle not in excludes:
                self.driver.switch_to.window(handle)
                return
        raise WindowNotFound('No window not matching excludes %s found.'
                             % excludes)

    def _select_matching(self, matcher, error):
        try:
            starting_handle = self.driver.current_window_handle
        except NoSuchWindowException:
            starting_handle = None
        for handle in self.driver.window_handles:
            self.driver.switch_to.window(handle)
            if matcher(self._get_current_window_info()):
                return
        if starting_handle:
            self.driver.switch_to.window(starting_handle)
        raise WindowNotFound(error)

    def _get_current_window_info(self):
        try:
            id, name = self.driver.execute_script(
                "return [ window.id, window.name ];")
        except WebDriverException:
            # The webdriver implementation doesn't support Javascript so we
            # can't get window id or name this way.
            id = name = None
        return WindowInfo(self.driver.current_window_handle,
                          id if id is not None else 'undefined',
                          name or 'undefined',
                          self.driver.title or 'undefined',
                          self.driver.current_url or 'undefined')
